program Oper

  use sep
  use fbi_mod
  use tomo_mod
  use separate_mod
  use mute_mod

  implicit none

  integer                             :: nt, n1, n2, wfld_j, gpu_id
  integer                             :: sou_nx, sou_ox, sou_dx, sou_z
  integer                             :: rec_nx, rec_ox, rec_dx, rec_z
  integer                             :: h_nx, h_ox, ih
  integer                             :: bc1, bc2, bc3, bc4
  real                                :: dt, d1, o1, d2, o2

  logical                             :: dso, mute
  real                                :: eps, dist1, dist2
  real, dimension(:), allocatable     :: wavelet, res
  real, dimension(:,:), allocatable   :: vel_init, vel0, slowsq0, gradient_slowsq
  real, dimension(:,:,:), allocatable :: data_obs, data, refl0, gradient_refl, gradient_data, hwt, model0, model1

  integer                             :: stat, iter, niter, itry, ntry, counter

  double precision                    :: r0, r1, r2, r3, step_size

  integer, dimension(8)               :: timer0, timer1
  real                                :: timer_sum=0.

  call sep_init()

!---------------------------------------------------------------------------
  call DATE_AND_TIME(values=timer0)
!---------------------------------------------------------------------------

  ! Read background velocity geometry
  call from_history("n1", n1)
  call from_history("o1", o1)
  call from_history("d1", d1)
  call from_history("n2", n2)
  call from_history("o2", o2)
  call from_history("d2", d2)

  ! Read wavelet information
  call from_aux("wavelet", "d1", dt)

  ! Read shots geometry
  call from_param("sou_nx", sou_nx)
  call from_param("sou_ox", sou_ox)
  call from_param("sou_dx", sou_dx)
  call from_param("sou_z", sou_z)
  ! Read receivers geometry
  call from_param("rec_nx", rec_nx)
  call from_param("rec_ox", rec_ox)
  call from_param("rec_dx", rec_dx)
  call from_param("rec_z", rec_z)

  if(exist_file("refl")) then
    call from_aux("refl", "n3", h_nx)
    call from_aux("refl", "o3", h_ox)
  else
    call from_param("h_nx", h_nx)
    call from_param("h_ox", h_ox)
  endif

  if(1-h_ox < 1 .or. 1-h_ox > h_nx) call erexit("h axis values are not correct")

  ! Read recording time
  call from_param("nt", nt)

  ! Read in Density
  call from_param("bc1", bc1)
  call from_param("bc2", bc2)
  call from_param("bc3", bc3)
  call from_param("bc4", bc4)
  call from_param("wfld_j", wfld_j)
  call from_param("gpu_id", gpu_id)

  call from_param('niter', niter, 10)
  call from_param('ntry', ntry, 2)
  call from_param('dso', dso)
  call from_param('mute', mute)

  call from_param('dist1', dist1)
  call from_param('dist2', dist2)

  if(dso) then
    call from_param('eps', eps)
    if(eps>1. .or. eps<0.) call erexit("eps value is not correct")
  endif

  call to_history("n1", n1)
  call to_history("o1", o1)
  call to_history("d1", d1)
  call to_history("n2", n2)
  call to_history("o2", o2)
  call to_history("d2", d2)

  call to_history("n1", n1, "modelr")
  call to_history("o1", o1, "modelr")
  call to_history("d1", d1, "modelr")
  call to_history("n2", n2, "modelr")
  call to_history("o2", o2, "modelr")
  call to_history("d2", d2, "modelr")
  call to_history("n3", h_nx, "modelr")
  call to_history("o3", h_ox, "modelr")
  call to_history("d3", 1., "modelr")

  call to_history("n1", n1, "mmovb")
  call to_history("o1", o1, "mmovb")
  call to_history("d1", d1, "mmovb")
  call to_history("n2", n2, "mmovb")
  call to_history("o2", o2, "mmovb")
  call to_history("d2", d2, "mmovb")

  call to_history("n1", n1, "mmovr")
  call to_history("o1", o1, "mmovr")
  call to_history("d1", d1, "mmovr")
  call to_history("n2", n2, "mmovr")
  call to_history("o2", o2, "mmovr")
  call to_history("d2", d2, "mmovr")
  call to_history("n3", h_nx, "mmovr")
  call to_history("o3", h_ox, "mmovr")
  call to_history("d3", 1., "mmovr")

  call to_history("n1", 2, "res")
  call to_history("o1", 0., "res")
  call to_history("d1", 1., "res")

  ! Allocation
  allocate( wavelet(nt))
  allocate( res(2))
  allocate( vel_init(n1, n2))
  allocate( vel0(n1, n2))
  allocate( slowsq0(n1, n2))
  allocate( gradient_slowsq(n1, n2))
  allocate( data(nt, rec_nx, sou_nx))
  allocate( data_obs(nt, rec_nx, sou_nx))
  allocate( refl0(n1, n2, h_nx))
  allocate( gradient_refl(n1, n2, h_nx))
  allocate( gradient_data(nt, rec_nx, sou_nx))
  allocate( model0(n1, n2, h_nx))
  allocate( model1(n1, n2, h_nx))
  if(dso) allocate( hwt(n1, n2, h_nx))

  wavelet = 0.
  vel0 = 0.
  data_obs = 0.
  refl0 = 0.
  res = 0.
  model0 = 0.
  model1 = 0.

  call sep_read(vel0)
  call sep_read(wavelet, "wavelet")
  call sep_read(data_obs, "data")
  if(exist_file("refl")) call sep_read(refl0, "refl")

  vel_init = vel0

  slowsq0 = 1./(vel0*vel0)

  ! Call L_initiate
  call init1(n1, o1, d1, n2, o2, d2, bc1, bc2, bc3, bc4)
  call init2(sou_nx, sou_ox, sou_dx, sou_z, rec_nx, rec_ox, rec_dx, rec_z)
  call init3(h_nx, h_ox, nt, dt, wfld_j, gpu_id)
  call init4(wavelet)

  if(dso) then
    do ih=1,h_nx
      hwt(:,:,ih) = ih-1+h_ox
    enddo
    hwt = abs(hwt)
    if(maxval(hwt)>0.) hwt = eps*hwt/maxval(hwt)
  endif

  if(mute) then
    call mute_init(nt, sou_nx, sou_ox, sou_dx, rec_nx, rec_ox, rec_dx)
    if(mute) call mute_offset(data_obs)
  endif

  call separate_init(n1, n2, d1, d2, dist1, dist2)

  write(0,*) " "

  if(maxval(abs(refl0)) .eq. 0.) then
    call model_update(vel0)
    data = - data_obs
    r0 = sum(dprod(data,data))
    write(0,*) "first refl residual =", real(r0)
    call L_oper2(.true., .false., gradient_refl, data)
    call L_oper2(.false., .false., gradient_refl, gradient_data)
    if(mute) call mute_offset(gradient_data)
    step_size = sum( dprod( gradient_data, data)) / sum( dprod( gradient_data, gradient_data))
    refl0 = refl0 - step_size * gradient_refl
    write(0,*) "initial step size is =", step_size
  endif

  model0 = refl0
  model0(:,:,1-h_ox) = model0(:,:,1-h_ox) + slowsq0
  call separate_smooth(model0(:,:,1-h_ox), slowsq0)
  do ih=1,h_nx
    call separate_rough(model0(:,:,1-h_ox), refl0(:,:,1-h_ox))
  enddo
  vel0 = 1./sqrt(slowsq0)

  counter = 0

  do iter=1,niter
    write(0,*) " "
    call model_update(vel0)
    call L_oper2(.false., .false., refl0, data)
    if(mute) call mute_offset(data)
    data = data - data_obs
    r0 = sum(dprod(data,data))

    call L_oper2(.true., .false., gradient_refl, data)

    call refl_update(refl0)
    call T_oper2(.true., .false., gradient_slowsq, data)
    write(0,*) "gs stats", minval(gradient_slowsq), maxval(gradient_slowsq), sum(gradient_slowsq)/size(gradient_slowsq), sqrt(sum(gradient_slowsq*gradient_slowsq))
    write(0,*) "gr stats", minval(gradient_refl), maxval(gradient_refl), sum(gradient_refl)/size(gradient_refl), sqrt(sum(gradient_refl*gradient_refl))
    write(0,*) " "

    do itry=1,2*ntry

      if(itry .eq. ntry .or. step_size .eq. 0.) then
        step_size = minval(slowsq0) / maxval(abs(gradient_slowsq)) / 1000.
        write(0,*) "Step size is resetted to ", step_size
      endif

      model1 = model0 - gradient_refl * step_size * 0.5
      model1(:,:,1-h_ox) = model1(:,:,1-h_ox) - gradient_slowsq * step_size * 0.5
      call separate_smooth(model1(:,:,1-h_ox), slowsq0)
      do ih=1,h_nx
        call separate_rough(model1(:,:,1-h_ox), refl0(:,:,1-h_ox))
      enddo
      vel0 = 1./sqrt(slowsq0)
      call model_update(vel0)
      call L_oper2(.false., .false., refl0, data)
      if(mute) call mute_offset(data)
      data = data - data_obs
      r1 = sum(dprod(data,data))

      model1 = model0 - gradient_refl * step_size * 1.0
      model1(:,:,1-h_ox) = model1(:,:,1-h_ox) - gradient_slowsq * step_size * 1.0
      call separate_smooth(model1(:,:,1-h_ox), slowsq0)
      do ih=1,h_nx
        call separate_rough(model1(:,:,1-h_ox), refl0(:,:,1-h_ox))
      enddo
      vel0 = 1./sqrt(slowsq0)
      call model_update(vel0)
      call L_oper2(.false., .false., refl0, data)
      if(mute) call mute_offset(data)
      data = data - data_obs
      r2 = sum(dprod(data,data))

      model1 = model0 - gradient_refl * step_size * 1.5
      model1(:,:,1-h_ox) = model1(:,:,1-h_ox) - gradient_slowsq * step_size * 1.5
      call separate_smooth(model1(:,:,1-h_ox), slowsq0)
      do ih=1,h_nx
        call separate_rough(model1(:,:,1-h_ox), refl0(:,:,1-h_ox))
      enddo
      vel0 = 1./sqrt(slowsq0)
      call model_update(vel0)
      call L_oper2(.false., .false., refl0, data)
      if(mute) call mute_offset(data)
      data = data - data_obs
      r3 = sum(dprod(data,data))

      write(0,*) "residuals =", real(r0), real(r1), real(r2), real(r3)
      call stepper(r0, r1, r2, r3, step_size, stat)

      if(stat .eq. 0) exit

    end do ! end of trying loop

    if(stat .ne. 0) then
      write(0,*) "trying iterations reached limit without finding a proper step size, will exit the program"
      exit
    endif

    model0 = model0 - gradient_refl * step_size
    model0(:,:,1-h_ox) = model0(:,:,1-h_ox) - gradient_slowsq * step_size
    call separate_smooth(model0(:,:,1-h_ox), slowsq0)
    do ih=1,h_nx
      call separate_rough(model0(:,:,1-h_ox), refl0(:,:,1-h_ox))
    enddo
    vel0 = 1./sqrt(slowsq0)

    res(1) = real(r0)
    if(dso) then
!      call hwindow(refl0, hwt, h_nx, h_ox, eps)
      res(2) = sum(refl0*refl0*hwt*hwt)
      refl0 = refl0 - refl0 * hwt * hwt

      model0 = refl0
      model0(:,:,1-h_ox) = model0(:,:,1-h_ox) + slowsq0
      call separate_smooth(model0(:,:,1-h_ox), slowsq0)
      do ih=1,h_nx
        call separate_rough(model0(:,:,1-h_ox), refl0(:,:,1-h_ox))
      enddo
      vel0 = 1./sqrt(slowsq0)
    endif

    ! Output model
    call sep_write(vel0-vel_init, "mmovb")
    call sep_write(refl0, "mmovr")
    call sep_write(res, "res")
    counter = counter + 1
    call to_history("n3", counter, "mmovb")
    call to_history("n4", counter, "mmovr")
    call to_history("n2", counter, "res")
  end do ! end of iteration loop

  call sep_write(vel0)
  call sep_write(refl0, "modelr")

!---------------------------------------------------------------------------
  call DATE_AND_TIME(values=timer1)
  timer0 = timer1-timer0
  timer_sum = timer_sum + (((timer0(3)*24+timer0(5))*60+timer0(6))*60+timer0(7))*1000+timer0(8)
!---------------------------------------------------------------------------
!---------------------------------------------------------------------------
  call to_history("timer sum (sec)", timer_sum/1000)
  call to_history("timer sum (min)", timer_sum/1000/60)
  write(0,*) "timer sum (sec)", timer_sum/1000
  write(0,*) "timer sum (min)", timer_sum/1000/60
  write(0,*) "DONE!"
!---------------------------------------------------------------------------

  call sep_close()

end program
