#include "cudaErrors.cu"

#define dloce(z,x) ((z)+(x)*n1e_gpu)
#define dlocebc(z,x) (((z)+bc1_gpu+FAT)+((x)+bc2_gpu+FAT)*n1e_gpu)

#define BLOCK_SIZE 16
#define FAT 5
#define COEFFS_SIZE 11
#define C0  0
#define CZ1 1
#define CX1 2
#define CZ2 3
#define CX2 4
#define CZ3 5
#define CX3 6
#define CZ4 7
#define CX4 8
#define CZ5 9
#define CX5 10

__constant__ int nt_gpu, n1_gpu, n2_gpu, n3_gpu;
__constant__ float dt_gpu, d1_gpu, d2_gpu;
__constant__ int h_nx_gpu, h_ox_gpu, n1e_gpu, n2e_gpu, bc1_gpu, bc2_gpu, wfld_j_gpu;
__constant__ int sou_nx_gpu, sou_ox_gpu, sou_dx_gpu, sou_z_gpu;
__constant__ int rec_nx_gpu, rec_ox_gpu, rec_dx_gpu, rec_z_gpu;
__constant__ float *p0_gpu, *p1_gpu, *p2_gpu, *p3_gpu, *p4_gpu, *p5_gpu, *p6_gpu, *p7_gpu, *p8_gpu, *p9_gpu;
__constant__ float *pinj0_gpu, *pinj1_gpu, *pinj2_gpu, *pinj3_gpu, *pinj4_gpu;
__constant__ float *source_gpu, *sourcesd_gpu, *sourcefd_gpu, *wfld1_gpu, *wfld2_gpu, *refl_gpu, *deltaB_gpu, *data_gpu;
__constant__ float coeffs[COEFFS_SIZE];

texture <float, 2, cudaReadModeElementType> velsq_ref, vwt_ref, wt_ref;
cudaChannelFormatDesc cf;
cudaArray *velsq_array, *vwt_array, *wt_array;

__global__ void wave_kernel(float *p0, float *p1, float *p2){
  __shared__ float p1s[BLOCK_SIZE+2*FAT][BLOCK_SIZE+2*FAT];
  register int ig=FAT+blockIdx.x*blockDim.x+threadIdx.x;   // global z coordinates
  register int jg=FAT+blockIdx.y*blockDim.y+threadIdx.y;   // global x coordinates
  register int il=FAT+threadIdx.x;               // thread-local z coordinate
  register int jl=FAT+threadIdx.y;               // thread-local x coordinate
  register int addr=dloce(ig,jg);                  // linear memroy address

  p1s[jl][il]=p1[addr];
  if(threadIdx.x<FAT) p1s[jl][il-FAT]=p1[dloce(ig-FAT,jg)];
  if(threadIdx.x>=BLOCK_SIZE-FAT) p1s[jl][il+FAT]=p1[dloce(ig+FAT,jg)];
  if(threadIdx.y<FAT) p1s[jl-FAT][il]=p1[dloce(ig,jg-FAT)];
  if(threadIdx.y>=BLOCK_SIZE-FAT) p1s[jl+FAT][il]=p1[dloce(ig,jg+FAT)];
  if(ig>=n1e_gpu-FAT || jg>=n2e_gpu-FAT) return;
  __syncthreads();

  p2[addr]= tex2D(velsq_ref, ig,jg)*(
      coeffs[CX5]*(p1s[jl-5][il]+p1s[jl+5][il])
      +coeffs[CZ5]*(p1s[jl][il-5]+p1s[jl][il+5])
      +coeffs[CX4]*(p1s[jl-4][il]+p1s[jl+4][il])
      +coeffs[CZ4]*(p1s[jl][il-4]+p1s[jl][il+4])
      +coeffs[CX3]*(p1s[jl-3][il]+p1s[jl+3][il])
      +coeffs[CZ3]*(p1s[jl][il-3]+p1s[jl][il+3])
      +coeffs[CX2]*(p1s[jl-2][il]+p1s[jl+2][il])
      +coeffs[CZ2]*(p1s[jl][il-2]+p1s[jl][il+2])
      +coeffs[CX1]*(p1s[jl-1][il]+p1s[jl+1][il])
      +coeffs[CZ1]*(p1s[jl][il-1]+p1s[jl][il+1])
      +coeffs[C0]*p1s[jl][il]
      )+p1s[jl][il]+p1s[jl][il]-p0[addr];

  if(ig<FAT+bc1_gpu) p2[addr]-= tex2D(wt_ref, ig,jg)*(p2[addr]-p1s[jl][il]+tex2D(vwt_ref, ig,jg)*(p1s[jl][il-1]-p1s[jl][il]));
  if(ig>=n1_gpu+FAT+bc1_gpu) p2[addr]-= tex2D(wt_ref, ig,jg)*(p2[addr]-p1s[jl][il]+tex2D(vwt_ref, ig,jg)*(p1s[jl][il+1]-p1s[jl][il]));
  if(jg<FAT+bc2_gpu) p2[addr]-= tex2D(wt_ref, ig,jg)*(p2[addr]-p1s[jl][il]+tex2D(vwt_ref, ig,jg)*(p1s[jl-1][il]-p1s[jl][il]));
  if(jg>=n2_gpu+FAT+bc2_gpu) p2[addr]-= tex2D(wt_ref, ig,jg)*(p2[addr]-p1s[jl][il]+tex2D(vwt_ref, ig,jg)*(p1s[jl+1][il]-p1s[jl][il]));

}

__global__ void wave_inj_kernel(float *p0, float *p1, float *p2, float *pinj){
  __shared__ float p1s[BLOCK_SIZE+2*FAT][BLOCK_SIZE+2*FAT];
  register int ig=FAT+blockIdx.x*blockDim.x+threadIdx.x;   // global z coordinates
  register int jg=FAT+blockIdx.y*blockDim.y+threadIdx.y;   // global x coordinates
  register int il=FAT+threadIdx.x;               // thread-local z coordinate
  register int jl=FAT+threadIdx.y;               // thread-local x coordinate
  register int addr=dloce(ig,jg);                  // linear memroy address

  p1s[jl][il]=p1[addr];
  if(threadIdx.x<FAT) p1s[jl][il-FAT]=p1[dloce(ig-FAT,jg)];
  if(threadIdx.x>=BLOCK_SIZE-FAT) p1s[jl][il+FAT]=p1[dloce(ig+FAT,jg)];
  if(threadIdx.y<FAT) p1s[jl-FAT][il]=p1[dloce(ig,jg-FAT)];
  if(threadIdx.y>=BLOCK_SIZE-FAT) p1s[jl+FAT][il]=p1[dloce(ig,jg+FAT)];
  if(ig>=n1e_gpu-FAT || jg>=n2e_gpu-FAT) return;
  __syncthreads();

  p2[addr]= tex2D(velsq_ref, ig,jg)*(pinj[addr]+
      coeffs[CX5]*(p1s[jl-5][il]+p1s[jl+5][il])
      +coeffs[CZ5]*(p1s[jl][il-5]+p1s[jl][il+5])
      +coeffs[CX4]*(p1s[jl-4][il]+p1s[jl+4][il])
      +coeffs[CZ4]*(p1s[jl][il-4]+p1s[jl][il+4])
      +coeffs[CX3]*(p1s[jl-3][il]+p1s[jl+3][il])
      +coeffs[CZ3]*(p1s[jl][il-3]+p1s[jl][il+3])
      +coeffs[CX2]*(p1s[jl-2][il]+p1s[jl+2][il])
      +coeffs[CZ2]*(p1s[jl][il-2]+p1s[jl][il+2])
      +coeffs[CX1]*(p1s[jl-1][il]+p1s[jl+1][il])
      +coeffs[CZ1]*(p1s[jl][il-1]+p1s[jl][il+1])
      +coeffs[C0]*p1s[jl][il]
      )+p1s[jl][il]+p1s[jl][il]-p0[addr];

  if(ig<FAT+bc1_gpu) p2[addr]-= tex2D(wt_ref, ig,jg)*(p2[addr]-p1s[jl][il]+tex2D(vwt_ref, ig,jg)*(p1s[jl][il-1]-p1s[jl][il]));
  if(ig>=n1_gpu+FAT+bc1_gpu) p2[addr]-= tex2D(wt_ref, ig,jg)*(p2[addr]-p1s[jl][il]+tex2D(vwt_ref, ig,jg)*(p1s[jl][il+1]-p1s[jl][il]));
  if(jg<FAT+bc2_gpu) p2[addr]-= tex2D(wt_ref, ig,jg)*(p2[addr]-p1s[jl][il]+tex2D(vwt_ref, ig,jg)*(p1s[jl-1][il]-p1s[jl][il]));
  if(jg>=n2_gpu+FAT+bc2_gpu) p2[addr]-= tex2D(wt_ref, ig,jg)*(p2[addr]-p1s[jl][il]+tex2D(vwt_ref, ig,jg)*(p1s[jl+1][il]-p1s[jl][il]));
}

__global__ void wave_adj_kernel(float *p0, float *p1, float *p2){
  __shared__ float p1s[BLOCK_SIZE+2*FAT][BLOCK_SIZE+2*FAT];
  register int ig=FAT+blockIdx.x*blockDim.x+threadIdx.x;   // global z coordinates
  register int jg=FAT+blockIdx.y*blockDim.y+threadIdx.y;   // global x coordinates
  register int il=FAT+threadIdx.x;               // thread-local z coordinate
  register int jl=FAT+threadIdx.y;               // thread-local x coordinate
  register int addr=dloce(ig,jg);                  // linear memroy address

  p1s[jl][il]=p1[addr];
  if(threadIdx.x<FAT) p1s[jl][il-FAT]=p1[dloce(ig-FAT,jg)];
  if(threadIdx.x>=BLOCK_SIZE-FAT) p1s[jl][il+FAT]=p1[dloce(ig+FAT,jg)];
  if(threadIdx.y<FAT) p1s[jl-FAT][il]=p1[dloce(ig,jg-FAT)];
  if(threadIdx.y>=BLOCK_SIZE-FAT) p1s[jl+FAT][il]=p1[dloce(ig,jg+FAT)];
  if(ig>=n1e_gpu-FAT || jg>=n2e_gpu-FAT) return;
  __syncthreads();

  p2[addr]=
      coeffs[CX5]*(p1s[jl-5][il]*tex2D(velsq_ref,ig,jg-5)
          +p1s[jl+5][il]*tex2D(velsq_ref,ig,jg+5))
      +coeffs[CZ5]*(p1s[jl][il-5]*tex2D(velsq_ref,ig-5,jg)
          +p1s[jl][il+5]*tex2D(velsq_ref,ig+5,jg))
      +coeffs[CX4]*(p1s[jl-4][il]*tex2D(velsq_ref,ig,jg-4)
          +p1s[jl+4][il]*tex2D(velsq_ref,ig,jg+4))
      +coeffs[CZ4]*(p1s[jl][il-4]*tex2D(velsq_ref,ig-4,jg)
          +p1s[jl][il+4]*tex2D(velsq_ref,ig+4,jg))
      +coeffs[CX3]*(p1s[jl-3][il]*tex2D(velsq_ref,ig,jg-3)
          +p1s[jl+3][il]*tex2D(velsq_ref,ig,jg+3))
      +coeffs[CZ3]*(p1s[jl][il-3]*tex2D(velsq_ref,ig-3,jg)
          +p1s[jl][il+3]*tex2D(velsq_ref,ig+3,jg))
      +coeffs[CX2]*(p1s[jl-2][il]*tex2D(velsq_ref,ig,jg-2)
          +p1s[jl+2][il]*tex2D(velsq_ref,ig,jg+2))
      +coeffs[CZ2]*(p1s[jl][il-2]*tex2D(velsq_ref,ig-2,jg)
          +p1s[jl][il+2]*tex2D(velsq_ref,ig+2,jg))
      +coeffs[CX1]*(p1s[jl-1][il]*tex2D(velsq_ref,ig,jg-1)
          +p1s[jl+1][il]*tex2D(velsq_ref,ig,jg+1))
      +coeffs[CZ1]*(p1s[jl][il-1]*tex2D(velsq_ref,ig-1,jg)
          +p1s[jl][il+1]*tex2D(velsq_ref,ig+1,jg))
      +coeffs[C0]*p1s[jl][il]*tex2D(velsq_ref,ig,jg)
      +p1s[jl][il]+p1s[jl][il]-p0[addr];

  if(ig<FAT+bc1_gpu) p2[addr]-= tex2D(wt_ref, ig,jg)*(p2[addr]-p1s[jl][il]+tex2D(vwt_ref, ig,jg)*(p1s[jl][il-1]-p1s[jl][il]));
  if(ig>=n1_gpu+FAT+bc1_gpu) p2[addr]-= tex2D(wt_ref, ig,jg)*(p2[addr]-p1s[jl][il]+tex2D(vwt_ref, ig,jg)*(p1s[jl][il+1]-p1s[jl][il]));
  if(jg<FAT+bc2_gpu) p2[addr]-= tex2D(wt_ref, ig,jg)*(p2[addr]-p1s[jl][il]+tex2D(vwt_ref, ig,jg)*(p1s[jl-1][il]-p1s[jl][il]));
  if(jg>=n2_gpu+FAT+bc2_gpu) p2[addr]-= tex2D(wt_ref, ig,jg)*(p2[addr]-p1s[jl][il]+tex2D(vwt_ref, ig,jg)*(p1s[jl+1][il]-p1s[jl][il]));
}

__global__ void wave_adj_inj_kernel(float *p0, float *p1, float *p2, float *pinj){
  __shared__ float p1s[BLOCK_SIZE+2*FAT][BLOCK_SIZE+2*FAT];
  register int ig=FAT+blockIdx.x*blockDim.x+threadIdx.x;   // global z coordinates
  register int jg=FAT+blockIdx.y*blockDim.y+threadIdx.y;   // global x coordinates
  register int il=FAT+threadIdx.x;               // thread-local z coordinate
  register int jl=FAT+threadIdx.y;               // thread-local x coordinate
  register int addr=dloce(ig,jg);                  // linear memroy address

  p1s[jl][il]=p1[addr];
  if(threadIdx.x<FAT) p1s[jl][il-FAT]=p1[dloce(ig-FAT,jg)];
  if(threadIdx.x>=BLOCK_SIZE-FAT) p1s[jl][il+FAT]=p1[dloce(ig+FAT,jg)];
  if(threadIdx.y<FAT) p1s[jl-FAT][il]=p1[dloce(ig,jg-FAT)];
  if(threadIdx.y>=BLOCK_SIZE-FAT) p1s[jl+FAT][il]=p1[dloce(ig,jg+FAT)];
  if(ig>=n1e_gpu-FAT || jg>=n2e_gpu-FAT) return;
  __syncthreads();

  p2[addr]=tex2D(velsq_ref, ig,jg)*pinj[addr]+
      coeffs[CX5]*(p1s[jl-5][il]*tex2D(velsq_ref,ig,jg-5)
          +p1s[jl+5][il]*tex2D(velsq_ref,ig,jg+5))
      +coeffs[CZ5]*(p1s[jl][il-5]*tex2D(velsq_ref,ig-5,jg)
          +p1s[jl][il+5]*tex2D(velsq_ref,ig+5,jg))
      +coeffs[CX4]*(p1s[jl-4][il]*tex2D(velsq_ref,ig,jg-4)
          +p1s[jl+4][il]*tex2D(velsq_ref,ig,jg+4))
      +coeffs[CZ4]*(p1s[jl][il-4]*tex2D(velsq_ref,ig-4,jg)
          +p1s[jl][il+4]*tex2D(velsq_ref,ig+4,jg))
      +coeffs[CX3]*(p1s[jl-3][il]*tex2D(velsq_ref,ig,jg-3)
          +p1s[jl+3][il]*tex2D(velsq_ref,ig,jg+3))
      +coeffs[CZ3]*(p1s[jl][il-3]*tex2D(velsq_ref,ig-3,jg)
          +p1s[jl][il+3]*tex2D(velsq_ref,ig+3,jg))
      +coeffs[CX2]*(p1s[jl-2][il]*tex2D(velsq_ref,ig,jg-2)
          +p1s[jl+2][il]*tex2D(velsq_ref,ig,jg+2))
      +coeffs[CZ2]*(p1s[jl][il-2]*tex2D(velsq_ref,ig-2,jg)
          +p1s[jl][il+2]*tex2D(velsq_ref,ig+2,jg))
      +coeffs[CX1]*(p1s[jl-1][il]*tex2D(velsq_ref,ig,jg-1)
          +p1s[jl+1][il]*tex2D(velsq_ref,ig,jg+1))
      +coeffs[CZ1]*(p1s[jl][il-1]*tex2D(velsq_ref,ig-1,jg)
          +p1s[jl][il+1]*tex2D(velsq_ref,ig+1,jg))
      +coeffs[C0]*p1s[jl][il]*tex2D(velsq_ref,ig,jg)
      +p1s[jl][il]+p1s[jl][il]-p0[addr];

  if(ig<FAT+bc1_gpu) p2[addr]-= tex2D(wt_ref, ig,jg)*(p2[addr]-p1s[jl][il]+tex2D(vwt_ref, ig,jg)*(p1s[jl][il-1]-p1s[jl][il]));
  if(ig>=n1_gpu+FAT+bc1_gpu) p2[addr]-= tex2D(wt_ref, ig,jg)*(p2[addr]-p1s[jl][il]+tex2D(vwt_ref, ig,jg)*(p1s[jl][il+1]-p1s[jl][il]));
  if(jg<FAT+bc2_gpu) p2[addr]-= tex2D(wt_ref, ig,jg)*(p2[addr]-p1s[jl][il]+tex2D(vwt_ref, ig,jg)*(p1s[jl-1][il]-p1s[jl][il]));
  if(jg>=n2_gpu+FAT+bc2_gpu) p2[addr]-= tex2D(wt_ref, ig,jg)*(p2[addr]-p1s[jl][il]+tex2D(vwt_ref, ig,jg)*(p1s[jl+1][il]-p1s[jl][il]));
}

__global__ void fat_update1_kernel(float *p0, float *p1)
{
  register int j=blockIdx.x*blockDim.x+threadIdx.x;
  register int i;

  if(j>=n2e_gpu) return;

  for(i=0; i<FAT; i++)
  {
    p1[dloce(i,j)] = 1./(1.+1.2*tex2D(vwt_ref,i,j))*((1.-1.2*tex2D(vwt_ref,i,j))*(p0[dloce(i,j)]-p1[dloce(i+1,j)]))+p0[dloce(i+1,j)];
  }
  for(i=n1e_gpu-1; i>=n1e_gpu-FAT; i--)
  {
    p1[dloce(i,j)] = 1./(1.+1.2*tex2D(vwt_ref,i,j))*((1.-1.2*tex2D(vwt_ref,i,j))*(p0[dloce(i,j)]-p1[dloce(i-1,j)]))+p0[dloce(i-1,j)];
  }
}

__global__ void fat_update2_kernel(float *p0, float *p1)
{
  register int i=blockIdx.x*blockDim.x+threadIdx.x;
  register int j;

  if(i>=n1e_gpu) return;

  for(j=0; j<FAT; j++)
  {
    p1[dloce(i,j)] = 1./(1.+1.2*tex2D(vwt_ref,i,j))*((1.-1.2*tex2D(vwt_ref,i,j))*(p0[dloce(i,j)]-p1[dloce(i,j+1)]))+p0[dloce(i,j+1)];
  }
  for(j=n2e_gpu-1; j>=n2e_gpu-FAT; j--)
  {
    p1[dloce(i,j)] = 1./(1.+1.2*tex2D(vwt_ref,i,j))*((1.-1.2*tex2D(vwt_ref,i,j))*(p0[dloce(i,j)]-p1[dloce(i,j-1)]))+p0[dloce(i,j-1)];
  }
}
__global__ void inject_source_kernel(float *source, float *pinj, int x, int its)
{
  register int i=blockIdx.x*blockDim.x+threadIdx.x;
  if(i>=1) return;
  pinj[dlocebc(sou_z_gpu,x)] = source[its];
}


__global__ void inject_data_kernel(float *data, float *pinj, int its)
{
  register int i=blockIdx.x*blockDim.x+threadIdx.x;
  if(i>=rec_nx_gpu) return;
  pinj[dlocebc(rec_z_gpu,i+rec_ox_gpu)] = data[its+i*nt_gpu];
}

__global__ void scat_kernel(float *p1, float *deltaB, float *pinj)
{
  register int i=blockIdx.x*blockDim.x+threadIdx.x;
  register int j=blockIdx.y*blockDim.y+threadIdx.y;

  if(i>=n1_gpu || j>=n2_gpu) return;
  pinj[dlocebc(i,j)] = p1[dlocebc(i,j)]*deltaB[i+(j)*n1_gpu];
}

__global__ void escat_for_kernel(float *p1, float *refl, float *pinj)
{
  register int i=blockIdx.x*blockDim.x+threadIdx.x;
  register int j=blockIdx.y*blockDim.y+threadIdx.y;
  register int ih, h_x;

  if(i>=n1_gpu || j>=n2_gpu) return;

  pinj[dlocebc(i,j)] = 0.;

  for(ih=0; ih<h_nx_gpu; ih++)
  {
    h_x = ih+h_ox_gpu;
    if(j-2*h_x>=0 && j-2*h_x<n2_gpu)
    {
      pinj[dlocebc(i,j)] += p1[dlocebc(i,j-2*h_x)]*refl[i+(j-h_x)*n1_gpu+ih*n1_gpu*n2_gpu];
    }
  }
}

__global__ void escat_adj_kernel(float *p1, float *refl, float *pinj)
{
  register int i=blockIdx.x*blockDim.x+threadIdx.x;
  register int j=blockIdx.y*blockDim.y+threadIdx.y;
  register int ih, h_x;

  if(i>=n1_gpu || j>=n2_gpu) return;

  pinj[dlocebc(i,j)] = 0.;

  for(ih=0; ih<h_nx_gpu; ih++)
  {
    h_x = ih+h_ox_gpu;
    if(j+2*h_x>=0 && j+2*h_x<n2_gpu)
    {
      pinj[dlocebc(i,j)] += p1[dlocebc(i,j+2*h_x)]*refl[i+(j+h_x)*n1_gpu+ih*n1_gpu*n2_gpu];
    }
  }
}

__global__ void image_kernel(float *p1, float *deltaB, float *wfld, int its)
{
  register int i=blockIdx.x*blockDim.x+threadIdx.x;
  register int j=blockIdx.y*blockDim.y+threadIdx.y;

  if(i>=n1_gpu || j>=n2_gpu) return;
  deltaB[i+j*n1_gpu] += wfld[i+j*n1_gpu+its*n1_gpu*n2_gpu]*p1[dlocebc(i,j)];
}

__global__ void eimage_kernel(float *p1, float *refl, float *wfld, int its)
{
  register int i=blockIdx.x*blockDim.x+threadIdx.x;
  register int j=blockIdx.y*blockDim.y+threadIdx.y;
  register int ih, h_x;

  if(i>=n1_gpu || j>=n2_gpu) return;

  for(ih=0; ih<h_nx_gpu; ih++)
  {
    h_x = ih+h_ox_gpu;
    if(j-abs(h_x)>=0 && j+abs(h_x)<n2_gpu)
    {
      refl[i+j*n1_gpu+ih*n1_gpu*n2_gpu] += wfld[i+(j-h_x)*n1_gpu+its*n1_gpu*n2_gpu]*p1[dlocebc(i,j+h_x)];
    }
  }
}

__global__ void collect_data_kernel(float *data, float *p, int its)
{
  register int i=blockIdx.x*blockDim.x+threadIdx.x;
  if(i>=rec_nx_gpu) return;
  data[its+i*nt_gpu] += p[dlocebc(rec_z_gpu,i+rec_ox_gpu)];
}

__global__ void collect_wfld_kernel(float *wfld, float *p, int its)
{
  register int i=blockIdx.x*blockDim.x+threadIdx.x;
  register int j=blockIdx.y*blockDim.y+threadIdx.y;
  if(i>=n1_gpu || j>=n2_gpu) return;
  wfld[i+j*n1_gpu+its*n1_gpu*n2_gpu] = p[dlocebc(i,j)];
}

extern "C" void gpu_setdevice(int *gpu_id)
{
  cudaSetDevice(*gpu_id);
}

extern "C" void gpu_init1(int *nt_in, int *n1_in, int *n2_in, int *n3_in, float *dt_in, float *d1_in, float *d2_in)
{
  int nt, n1, n2, n3;
  float dt, d1, d2;

  nt = (*nt_in);
  n1 = (*n1_in);
  n2 = (*n2_in);
  n3 = (*n3_in);
  dt = (*dt_in);
  d1 = (*d1_in);
  d2 = (*d2_in);

  cuda_call(cudaMemcpyToSymbol(nt_gpu,  &nt,  sizeof(int)));
  cuda_call(cudaMemcpyToSymbol(n1_gpu,  &n1,  sizeof(int)));
  cuda_call(cudaMemcpyToSymbol(n2_gpu,  &n2,  sizeof(int)));
  cuda_call(cudaMemcpyToSymbol(n3_gpu,  &n3,  sizeof(int)));
  cuda_call(cudaMemcpyToSymbol(dt_gpu,  &dt,  sizeof(float)));
  cuda_call(cudaMemcpyToSymbol(d1_gpu,  &d1,  sizeof(float)));
  cuda_call(cudaMemcpyToSymbol(d2_gpu,  &d2,  sizeof(float)));
}

extern "C" void gpu_init2(int *h_nx_in, int *h_ox_in, int *n1e_in, int *n2e_in, int *bc1_in, int *bc2_in, int *wfld_j_in)
{
  int h_nx, h_ox, n1e, n2e, bc1, bc2, wfld_j;

  h_nx = (*h_nx_in);
  h_ox = (*h_ox_in);
  n1e = (*n1e_in);
  n2e = (*n2e_in);
  bc1 = (*bc1_in);
  bc2 = (*bc2_in);
  wfld_j = (*wfld_j_in);

  cuda_call(cudaMemcpyToSymbol(h_nx_gpu, &h_nx, sizeof(int)));
  cuda_call(cudaMemcpyToSymbol(h_ox_gpu, &h_ox, sizeof(int)));
  cuda_call(cudaMemcpyToSymbol(n1e_gpu, &n1e, sizeof(int)));
  cuda_call(cudaMemcpyToSymbol(n2e_gpu, &n2e, sizeof(int)));
  cuda_call(cudaMemcpyToSymbol(bc1_gpu, &bc1, sizeof(int)));
  cuda_call(cudaMemcpyToSymbol(bc2_gpu, &bc2, sizeof(int)));
  cuda_call(cudaMemcpyToSymbol(wfld_j_gpu, &wfld_j, sizeof(int)));
}

extern "C" void gpu_init3(int *sou_nx_in, int *sou_ox_in, int *sou_dx_in, int *sou_z_in, int *rec_nx_in, int *rec_ox_in, int *rec_dx_in, int *rec_z_in)
{
  int sou_nx, sou_ox, sou_dx, sou_z;
  int rec_nx, rec_ox, rec_dx, rec_z;

  sou_nx = (*sou_nx_in);
  sou_ox = (*sou_ox_in);
  sou_dx = (*sou_dx_in);
  sou_z  = (*sou_z_in);
  rec_nx = (*rec_nx_in);
  rec_ox = (*rec_ox_in);
  rec_dx = (*rec_dx_in);
  rec_z  = (*rec_z_in);

  cuda_call(cudaMemcpyToSymbol(sou_nx_gpu, &sou_nx, sizeof(int)));
  cuda_call(cudaMemcpyToSymbol(sou_ox_gpu, &sou_ox, sizeof(int)));
  cuda_call(cudaMemcpyToSymbol(sou_dx_gpu, &sou_dx, sizeof(int)));
  cuda_call(cudaMemcpyToSymbol(sou_z_gpu, &sou_z, sizeof(int)));
  cuda_call(cudaMemcpyToSymbol(rec_nx_gpu, &rec_nx, sizeof(int)));
  cuda_call(cudaMemcpyToSymbol(rec_ox_gpu, &rec_ox, sizeof(int)));
  cuda_call(cudaMemcpyToSymbol(rec_dx_gpu, &rec_dx, sizeof(int)));
  cuda_call(cudaMemcpyToSymbol(rec_z_gpu, &rec_z, sizeof(int)));
}

extern "C" void gpu_alloc(float *source_cpu, float *sourcesd_cpu, float *sourcefd_cpu)
{
  int nt, n1, n2;
  float d1, d2;
  int h_nx, n1e, n2e, rec_nx;
  float *p0, *p1, *p2, *p3, *p4, *p5, *p6, *p7, *p8, *p9;
  float *pinj0, *pinj1, *pinj2, *pinj3, *pinj4;
  float *source, *sourcesd, *sourcefd, *wfld1, *wfld2, *refl, *deltaB, *data;
  int memsize;

  cuda_call(cudaMemcpyFromSymbol(&nt, nt_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&n1, n1_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&n2, n2_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&d1, d1_gpu, sizeof(float)));
  cuda_call(cudaMemcpyFromSymbol(&d2, d2_gpu, sizeof(float)));
  cuda_call(cudaMemcpyFromSymbol(&h_nx, h_nx_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&n1e, n1e_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&n2e, n2e_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&rec_nx, rec_nx_gpu, sizeof(int)));

  cuda_call(cudaMalloc( (void**) &p0,       n1e*n2e*sizeof(float)));
  cuda_call(cudaMalloc( (void**) &p1,       n1e*n2e*sizeof(float)));
  cuda_call(cudaMalloc( (void**) &p2,       n1e*n2e*sizeof(float)));
  cuda_call(cudaMalloc( (void**) &p3,       n1e*n2e*sizeof(float)));
  cuda_call(cudaMalloc( (void**) &p4,       n1e*n2e*sizeof(float)));
  cuda_call(cudaMalloc( (void**) &p5,       n1e*n2e*sizeof(float)));
  cuda_call(cudaMalloc( (void**) &p6,       n1e*n2e*sizeof(float)));
  cuda_call(cudaMalloc( (void**) &p7,       n1e*n2e*sizeof(float)));
  cuda_call(cudaMalloc( (void**) &p8,       n1e*n2e*sizeof(float)));
  cuda_call(cudaMalloc( (void**) &p9,       n1e*n2e*sizeof(float)));
  cuda_call(cudaMalloc( (void**) &pinj0,    n1e*n2e*sizeof(float)));
  cuda_call(cudaMalloc( (void**) &pinj1,    n1e*n2e*sizeof(float)));
  cuda_call(cudaMalloc( (void**) &pinj2,    n1e*n2e*sizeof(float)));
  cuda_call(cudaMalloc( (void**) &pinj3,    n1e*n2e*sizeof(float)));
  cuda_call(cudaMalloc( (void**) &pinj4,    n1e*n2e*sizeof(float)));
  cuda_call(cudaMalloc( (void**) &source,   nt*sizeof(float)));
  cuda_call(cudaMalloc( (void**) &sourcesd, nt*sizeof(float)));
  cuda_call(cudaMalloc( (void**) &sourcefd, nt*sizeof(float)));
  cuda_call(cudaMalloc( (void**) &wfld1,    n1*n2*nt*sizeof(float)));
  cuda_call(cudaMalloc( (void**) &wfld2,    n1*n2*nt*sizeof(float)));
  cuda_call(cudaMalloc( (void**) &refl,     n1*n2*h_nx*sizeof(float)));
  cuda_call(cudaMalloc( (void**) &deltaB,   n1*n2*sizeof(float)));
  cuda_call(cudaMalloc( (void**) &data,     nt*rec_nx*sizeof(float)));

  cuda_call(cudaMemcpyToSymbol(p0_gpu, &p0, sizeof(float*)));
  cuda_call(cudaMemcpyToSymbol(p1_gpu, &p1, sizeof(float*)));
  cuda_call(cudaMemcpyToSymbol(p2_gpu, &p2, sizeof(float*)));
  cuda_call(cudaMemcpyToSymbol(p3_gpu, &p3, sizeof(float*)));
  cuda_call(cudaMemcpyToSymbol(p4_gpu, &p4, sizeof(float*)));
  cuda_call(cudaMemcpyToSymbol(p5_gpu, &p5, sizeof(float*)));
  cuda_call(cudaMemcpyToSymbol(p6_gpu, &p6, sizeof(float*)));
  cuda_call(cudaMemcpyToSymbol(p7_gpu, &p7, sizeof(float*)));
  cuda_call(cudaMemcpyToSymbol(p8_gpu, &p8, sizeof(float*)));
  cuda_call(cudaMemcpyToSymbol(p9_gpu, &p9, sizeof(float*)));
  cuda_call(cudaMemcpyToSymbol(pinj0_gpu, &pinj0, sizeof(float*)));
  cuda_call(cudaMemcpyToSymbol(pinj1_gpu, &pinj1, sizeof(float*)));
  cuda_call(cudaMemcpyToSymbol(pinj2_gpu, &pinj2, sizeof(float*)));
  cuda_call(cudaMemcpyToSymbol(pinj3_gpu, &pinj3, sizeof(float*)));
  cuda_call(cudaMemcpyToSymbol(pinj4_gpu, &pinj4, sizeof(float*)));
  cuda_call(cudaMemcpyToSymbol(source_gpu, &source, sizeof(float*)));
  cuda_call(cudaMemcpyToSymbol(sourcesd_gpu, &sourcesd, sizeof(float*)));
  cuda_call(cudaMemcpyToSymbol(sourcefd_gpu, &sourcefd, sizeof(float*)));
  cuda_call(cudaMemcpyToSymbol(wfld1_gpu, &wfld1, sizeof(float*)));
  cuda_call(cudaMemcpyToSymbol(wfld2_gpu, &wfld2, sizeof(float*)));
  cuda_call(cudaMemcpyToSymbol(refl_gpu, &refl, sizeof(float*)));
  cuda_call(cudaMemcpyToSymbol(deltaB_gpu, &deltaB, sizeof(float*)));
  cuda_call(cudaMemcpyToSymbol(data_gpu, &data, sizeof(float*)));

  cf = cudaCreateChannelDesc<float>();
  cuda_call(cudaMallocArray(&velsq_array, &cf, n1e, n2e));
  cuda_call(cudaMallocArray(&vwt_array,  &cf, n1e, n2e));
  cuda_call(cudaMallocArray(&wt_array,   &cf, n1e, n2e));

  memsize = 18*n1e*n2e + 3*nt + 2*n1*n2*nt + n1*n2*h_nx + n1*n2 + nt*rec_nx;

  fprintf(stderr, " GPU memory allocated %f MB (%f GB) \n", memsize*4./1024/1024, memsize*4./1024/1024/1024);

  float coeffs_cpu[COEFFS_SIZE];
  coeffs_cpu[CZ1]= 3.333333333/(2.*d1*d1);
  coeffs_cpu[CX1]= 3.333333333/(2.*d2*d2);
  coeffs_cpu[CZ2]=-.4761904762/(2.*d1*d1);
  coeffs_cpu[CX2]=-.4761904762/(2.*d2*d2);
  coeffs_cpu[CZ3]= .0793650794/(2.*d1*d1);
  coeffs_cpu[CX3]= .0793650794/(2.*d2*d2);
  coeffs_cpu[CZ4]=-.0099206349/(2.*d1*d1);
  coeffs_cpu[CX4]=-.0099206349/(2.*d2*d2);
  coeffs_cpu[CZ5]= .0006349206/(2.*d1*d1);
  coeffs_cpu[CX5]= .0006349206/(2.*d2*d2);
  coeffs_cpu[C0]=coeffs_cpu[CZ5]+coeffs_cpu[CX5];
  coeffs_cpu[C0]+=coeffs_cpu[CZ4]+coeffs_cpu[CX4];
  coeffs_cpu[C0]+=coeffs_cpu[CZ3]+coeffs_cpu[CX3];
  coeffs_cpu[C0]+=coeffs_cpu[CZ2]+coeffs_cpu[CX2];
  coeffs_cpu[C0]+=coeffs_cpu[CZ1]+coeffs_cpu[CX1];
  coeffs_cpu[C0]=-2.*coeffs_cpu[C0];

  cuda_call(cudaMemcpyToSymbol(coeffs, coeffs_cpu, COEFFS_SIZE*sizeof(float)));

  cuda_call(cudaMemcpy(source, source_cpu, nt*sizeof(float),      cudaMemcpyHostToDevice));
  cuda_call(cudaMemcpy(sourcesd, sourcesd_cpu, nt*sizeof(float),      cudaMemcpyHostToDevice));
  cuda_call(cudaMemcpy(sourcefd, sourcefd_cpu, nt*sizeof(float),      cudaMemcpyHostToDevice));
}

extern "C" void gpu_model_update(float *velsq_cpu, float *vel_cpu, float *wt_cpu)
{
  int n1e, n2e;

  cuda_call(cudaMemcpyFromSymbol(&n1e, n1e_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&n2e, n2e_gpu, sizeof(int)));

  cuda_call(cudaMemcpyToArray(velsq_array, 0,0, velsq_cpu, n1e*n2e*sizeof(float), cudaMemcpyHostToDevice));
  cuda_call(cudaMemcpyToArray(vwt_array,  0,0, vel_cpu,  n1e*n2e*sizeof(float), cudaMemcpyHostToDevice));
  cuda_call(cudaMemcpyToArray(wt_array,   0,0, wt_cpu,   n1e*n2e*sizeof(float), cudaMemcpyHostToDevice));

  cuda_call(cudaBindTextureToArray(velsq_ref, velsq_array));
  cuda_call(cudaBindTextureToArray(vwt_ref,  vwt_array));
  cuda_call(cudaBindTextureToArray(wt_ref,   wt_array));
}

extern "C" void gpu_refl_update(float *refl_cpu)
{
  int n1, n2, h_nx;
  float *refl;

  cuda_call(cudaMemcpyFromSymbol(&n1, n1_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&n2, n2_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&h_nx, h_nx_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&refl, refl_gpu, sizeof(float*)));

  cuda_call(cudaMemcpy(refl, refl_cpu, n1*n2*h_nx*sizeof(float), cudaMemcpyHostToDevice));
}

extern "C" void gpu_nlmodel(int *tid, int *nth, float *data_cpu)
{
  int nt, n1, n2, n3;
  int n1e, n2e, wfld_j;
  int sou_nx, sou_ox, sou_dx, rec_nx;
  float *p0, *p1, *source, *data;
  float *pinj0;
  int sou_x, its;
  float *ptemp;

  cuda_call(cudaMemcpyFromSymbol(&nt, nt_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&n1, n1_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&n2, n2_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&n3, n3_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&n1e, n1e_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&n2e, n2e_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&wfld_j, wfld_j_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&sou_nx, sou_nx_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&sou_ox, sou_ox_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&sou_dx, sou_dx_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&rec_nx, rec_nx_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&p0, p0_gpu, sizeof(float*)));
  cuda_call(cudaMemcpyFromSymbol(&p1, p1_gpu, sizeof(float*)));
  cuda_call(cudaMemcpyFromSymbol(&pinj0, pinj0_gpu, sizeof(float*)));
  cuda_call(cudaMemcpyFromSymbol(&source, source_gpu, sizeof(float*)));
  cuda_call(cudaMemcpyFromSymbol(&data, data_gpu, sizeof(float*)));

  int nblocksx=ceil(1.*(n1e-2*FAT)/BLOCK_SIZE);
  int nblocksy=ceil(1.*(n2e-2*FAT)/BLOCK_SIZE);
  int nblocksx2=ceil(1.*n1/BLOCK_SIZE);
  int nblocksy2=ceil(1.*n2/BLOCK_SIZE);
  dim3 dimBlock(BLOCK_SIZE, BLOCK_SIZE);
  dim3 dimGrid(nblocksx,nblocksy);
  dim3 dimGrid2(nblocksx2,nblocksy2);

  for(int isou=(*tid); isou<sou_nx; isou+=(*nth))
  {
    sou_x=isou*sou_dx+sou_ox;

    cuda_call(cudaMemset(p0,   0,n1e*n2e*sizeof(float)));
    cuda_call(cudaMemset(p1,   0,n1e*n2e*sizeof(float)));
    cuda_call(cudaMemset(pinj0, 0,n1e*n2e*sizeof(float)));
    cuda_call(cudaMemset(data, 0,nt*rec_nx*sizeof(float)));

    its = 0;
    for(int it=0; it<n3; it++)
    {
      if(it%wfld_j==0)
      {
        kernel_exec(inject_source_kernel<<<1,1>>>(source, pinj0, sou_x, its));
        kernel_exec(wave_inj_kernel<<<dimGrid, dimBlock>>>(p0, p1, p0, pinj0));
      }
      else
      {
        kernel_exec(wave_kernel<<<dimGrid, dimBlock>>>(p0, p1, p0));
      }
      kernel_exec(fat_update1_kernel<<<n2e/80+1,80>>>(p1, p0));
      kernel_exec(fat_update2_kernel<<<n1e/80+1,80>>>(p1, p0));
      if(it%wfld_j==0)
      {
        kernel_exec(collect_data_kernel<<<rec_nx/80+1,80>>>(data, p1, its));
        its++;
      }

      ptemp=p0; p0=p1; p1=ptemp;
    }
    cuda_call(cudaMemcpy(&data_cpu[nt*rec_nx*isou], data, nt*rec_nx*sizeof(float), cudaMemcpyDeviceToHost));
  }
}

extern "C" void gpu_model(int *tid, int *nth, float *refl_cpu, float *data_cpu)
{
  int nt, n1, n2, n3;
  int h_nx, n1e, n2e, wfld_j;
  int sou_nx, sou_ox, sou_dx, rec_nx;
  float *p0, *p1, *p2, *p3, *sourcesd, *refl, *data;
  float *pinj0, *pinj1;
  int sou_x, its;
  float *ptemp;

  cuda_call(cudaMemcpyFromSymbol(&nt, nt_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&n1, n1_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&n2, n2_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&n3, n3_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&h_nx, h_nx_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&n1e, n1e_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&n2e, n2e_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&wfld_j, wfld_j_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&sou_nx, sou_nx_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&sou_ox, sou_ox_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&sou_dx, sou_dx_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&rec_nx, rec_nx_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&p0, p0_gpu, sizeof(float*)));
  cuda_call(cudaMemcpyFromSymbol(&p1, p1_gpu, sizeof(float*)));
  cuda_call(cudaMemcpyFromSymbol(&p2, p2_gpu, sizeof(float*)));
  cuda_call(cudaMemcpyFromSymbol(&p3, p3_gpu, sizeof(float*)));
  cuda_call(cudaMemcpyFromSymbol(&pinj0, pinj0_gpu, sizeof(float*)));
  cuda_call(cudaMemcpyFromSymbol(&pinj1, pinj1_gpu, sizeof(float*)));
  cuda_call(cudaMemcpyFromSymbol(&sourcesd, sourcesd_gpu, sizeof(float*)));
  cuda_call(cudaMemcpyFromSymbol(&refl, refl_gpu, sizeof(float*)));
  cuda_call(cudaMemcpyFromSymbol(&data, data_gpu, sizeof(float*)));

  cuda_call(cudaMemcpy(refl, refl_cpu, n1*n2*h_nx*sizeof(float), cudaMemcpyHostToDevice));

  int nblocksx=ceil(1.*(n1e-2*FAT)/BLOCK_SIZE);
  int nblocksy=ceil(1.*(n2e-2*FAT)/BLOCK_SIZE);
  int nblocksx2=ceil(1.*n1/BLOCK_SIZE);
  int nblocksy2=ceil(1.*n2/BLOCK_SIZE);
  dim3 dimBlock(BLOCK_SIZE, BLOCK_SIZE);
  dim3 dimGrid(nblocksx,nblocksy);
  dim3 dimGrid2(nblocksx2,nblocksy2);

  for(int isou=(*tid); isou<sou_nx; isou+=(*nth))
  {
    sou_x=isou*sou_dx+sou_ox;

    cuda_call(cudaMemset(p0,   0,n1e*n2e*sizeof(float)));
    cuda_call(cudaMemset(p1,   0,n1e*n2e*sizeof(float)));
    cuda_call(cudaMemset(p2,   0,n1e*n2e*sizeof(float)));
    cuda_call(cudaMemset(p3,   0,n1e*n2e*sizeof(float)));
    cuda_call(cudaMemset(pinj0, 0,n1e*n2e*sizeof(float)));
    cuda_call(cudaMemset(pinj1, 0,n1e*n2e*sizeof(float)));
    cuda_call(cudaMemset(data, 0,nt*rec_nx*sizeof(float)));

    its = 0;
    for(int it=0; it<n3; it++)
    {
      if(it%wfld_j==0)
      {
        kernel_exec(inject_source_kernel<<<1,1>>>(sourcesd, pinj0, sou_x, its));
        kernel_exec(escat_for_kernel<<<dimGrid2, dimBlock>>>(p1, refl, pinj1));
        kernel_exec(wave_inj_kernel<<<dimGrid, dimBlock>>>(p0, p1, p0, pinj0));
        kernel_exec(wave_inj_kernel<<<dimGrid, dimBlock>>>(p2, p3, p2, pinj1));
      }
      else
      {
        kernel_exec(wave_kernel<<<dimGrid, dimBlock>>>(p0, p1, p0));
        kernel_exec(wave_kernel<<<dimGrid, dimBlock>>>(p2, p3, p2));
      }
      kernel_exec(fat_update1_kernel<<<n2e/80+1,80>>>(p1, p0));
      kernel_exec(fat_update2_kernel<<<n1e/80+1,80>>>(p1, p0));
      kernel_exec(fat_update1_kernel<<<n2e/80+1,80>>>(p3, p2));
      kernel_exec(fat_update2_kernel<<<n1e/80+1,80>>>(p3, p2));
      if(it%wfld_j==0)
      {
        kernel_exec(collect_data_kernel<<<rec_nx/80+1,80>>>(data, p3, its));
        its++;
      }

      ptemp=p0; p0=p1; p1=ptemp;
      ptemp=p2; p2=p3; p3=ptemp;

    }
    cuda_call(cudaMemcpy(&data_cpu[nt*rec_nx*isou], data, nt*rec_nx*sizeof(float), cudaMemcpyDeviceToHost));
  }
}


extern "C" void gpu_migrate(int *tid, int *nth, float *refl_cpu, float *data_cpu)
{
  int nt, n1, n2, n3;
  int h_nx, n1e, n2e, wfld_j;
  int sou_nx, sou_ox, sou_dx, rec_nx;
  float *p0, *p1, *sourcesd, *wfld1, *refl, *data;
  float *pinj0;
  int sou_x, its;
  float *ptemp;

  cuda_call(cudaMemcpyFromSymbol(&nt, nt_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&n1, n1_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&n2, n2_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&n3, n3_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&h_nx, h_nx_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&n1e, n1e_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&n2e, n2e_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&wfld_j, wfld_j_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&sou_nx, sou_nx_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&sou_ox, sou_ox_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&sou_dx, sou_dx_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&rec_nx, rec_nx_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&p0, p0_gpu, sizeof(float*)));
  cuda_call(cudaMemcpyFromSymbol(&p1, p1_gpu, sizeof(float*)));
  cuda_call(cudaMemcpyFromSymbol(&pinj0, pinj0_gpu, sizeof(float*)));
  cuda_call(cudaMemcpyFromSymbol(&sourcesd, sourcesd_gpu, sizeof(float*)));
  cuda_call(cudaMemcpyFromSymbol(&wfld1, wfld1_gpu, sizeof(float*)));
  cuda_call(cudaMemcpyFromSymbol(&refl, refl_gpu, sizeof(float*)));
  cuda_call(cudaMemcpyFromSymbol(&data, data_gpu, sizeof(float*)));

  int nblocksx=ceil(1.*(n1e-2*FAT)/BLOCK_SIZE);
  int nblocksy=ceil(1.*(n2e-2*FAT)/BLOCK_SIZE);
  int nblocksx2=ceil(1.*n1/BLOCK_SIZE);
  int nblocksy2=ceil(1.*n2/BLOCK_SIZE);
  dim3 dimBlock(BLOCK_SIZE, BLOCK_SIZE);
  dim3 dimGrid(nblocksx,nblocksy);
  dim3 dimGrid2(nblocksx2,nblocksy2);

  cuda_call(cudaMemset(refl,   0,n1*n2*h_nx*sizeof(float)));

  for(int isou=(*tid); isou<sou_nx; isou+=(*nth))
  {
    sou_x=isou*sou_dx+sou_ox;

    cuda_call(cudaMemset(p0,   0,n1e*n2e*sizeof(float)));
    cuda_call(cudaMemset(p1,   0,n1e*n2e*sizeof(float)));
    cuda_call(cudaMemset(pinj0, 0,n1e*n2e*sizeof(float)));
    cuda_call(cudaMemset(wfld1, 0,n1*n2*nt*sizeof(float)));
    cuda_call(cudaMemset(data, 0,nt*rec_nx*sizeof(float)));

    cuda_call(cudaMemcpy(data, &data_cpu[nt*rec_nx*isou], nt*rec_nx*sizeof(float), cudaMemcpyHostToDevice));

    its = 0;
    for(int it=0; it<n3; it++)
    {
      if(it%wfld_j==0)
      {
        kernel_exec(inject_source_kernel<<<1,1>>>(sourcesd, pinj0, sou_x, its));
        kernel_exec(wave_inj_kernel<<<dimGrid, dimBlock>>>(p0, p1, p0, pinj0));
      }
      else
      {
        kernel_exec(wave_kernel<<<dimGrid, dimBlock>>>(p0, p1, p0));
      }
      kernel_exec(fat_update1_kernel<<<n2e/80+1,80>>>(p1, p0));
      kernel_exec(fat_update2_kernel<<<n1e/80+1,80>>>(p1, p0));
      if(it%wfld_j==0)
      {
        kernel_exec(collect_wfld_kernel<<<dimGrid2, dimBlock>>>(wfld1, p1, its));
        its++;
      }
      ptemp=p0; p0=p1; p1=ptemp;
    }

    cuda_call(cudaMemset(p0,   0,n1e*n2e*sizeof(float)));
    cuda_call(cudaMemset(p1,   0,n1e*n2e*sizeof(float)));
    cuda_call(cudaMemset(pinj0, 0,n1e*n2e*sizeof(float)));

    its = nt-1;
    for(int it=n3-1; it>=0; it--)
    {
      if(it%wfld_j==0)
      {
        kernel_exec(inject_data_kernel<<<rec_nx/80+1,80>>>(data, pinj0, its));
        kernel_exec(wave_adj_inj_kernel<<<dimGrid, dimBlock>>>(p0, p1, p0, pinj0));
      }
      else
      {
        kernel_exec(wave_adj_kernel<<<dimGrid, dimBlock>>>(p0, p1, p0));
      }
      kernel_exec(fat_update1_kernel<<<n2e/80+1,80>>>(p1, p0));
      kernel_exec(fat_update2_kernel<<<n1e/80+1,80>>>(p1, p0));
      if(it%wfld_j==0)
      {
        kernel_exec(eimage_kernel<<<dimGrid2, dimBlock>>>(p1, refl, wfld1, its));
        its--;
      }
      ptemp=p0; p0=p1; p1=ptemp;
    }
  }
  cuda_call(cudaMemcpy(refl_cpu, refl, n1*n2*h_nx*sizeof(float), cudaMemcpyDeviceToHost));
}

extern "C" void gpu_tomofor(int *tid, int *nth, float *deltaB_cpu, float *data_cpu)
{
  int nt, n1, n2, n3;
  int h_nx, n1e, n2e, wfld_j;
  int sou_nx, sou_ox, sou_dx, rec_nx;
  float *p0, *p1, *p2, *p3, *p4, *p5, *p6, *p7, *p8, *p9;
  float *pinj0, *pinj1, *pinj2, *pinj3, *pinj4;
  float *sourcefd, *refl, *deltaB, *data;
  int sou_x, its;
  float *ptemp;

  cuda_call(cudaMemcpyFromSymbol(&nt, nt_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&n1, n1_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&n2, n2_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&n3, n3_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&h_nx, h_nx_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&n1e, n1e_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&n2e, n2e_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&wfld_j, wfld_j_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&sou_nx, sou_nx_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&sou_ox, sou_ox_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&sou_dx, sou_dx_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&rec_nx, rec_nx_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&p0, p0_gpu, sizeof(float*)));
  cuda_call(cudaMemcpyFromSymbol(&p1, p1_gpu, sizeof(float*)));
  cuda_call(cudaMemcpyFromSymbol(&p2, p2_gpu, sizeof(float*)));
  cuda_call(cudaMemcpyFromSymbol(&p3, p3_gpu, sizeof(float*)));
  cuda_call(cudaMemcpyFromSymbol(&p4, p4_gpu, sizeof(float*)));
  cuda_call(cudaMemcpyFromSymbol(&p5, p5_gpu, sizeof(float*)));
  cuda_call(cudaMemcpyFromSymbol(&p6, p6_gpu, sizeof(float*)));
  cuda_call(cudaMemcpyFromSymbol(&p7, p7_gpu, sizeof(float*)));
  cuda_call(cudaMemcpyFromSymbol(&p8, p8_gpu, sizeof(float*)));
  cuda_call(cudaMemcpyFromSymbol(&p9, p9_gpu, sizeof(float*)));
  cuda_call(cudaMemcpyFromSymbol(&pinj0, pinj0_gpu, sizeof(float*)));
  cuda_call(cudaMemcpyFromSymbol(&pinj1, pinj1_gpu, sizeof(float*)));
  cuda_call(cudaMemcpyFromSymbol(&pinj2, pinj2_gpu, sizeof(float*)));
  cuda_call(cudaMemcpyFromSymbol(&pinj3, pinj3_gpu, sizeof(float*)));
  cuda_call(cudaMemcpyFromSymbol(&pinj4, pinj4_gpu, sizeof(float*)));
  cuda_call(cudaMemcpyFromSymbol(&sourcefd, sourcefd_gpu, sizeof(float*)));
  cuda_call(cudaMemcpyFromSymbol(&refl, refl_gpu, sizeof(float*)));
  cuda_call(cudaMemcpyFromSymbol(&deltaB, deltaB_gpu, sizeof(float*)));
  cuda_call(cudaMemcpyFromSymbol(&data, data_gpu, sizeof(float*)));

  cuda_call(cudaMemcpy(deltaB, deltaB_cpu, n1*n2*sizeof(float), cudaMemcpyHostToDevice));

  int nblocksx=ceil(1.*(n1e-2*FAT)/BLOCK_SIZE);
  int nblocksy=ceil(1.*(n2e-2*FAT)/BLOCK_SIZE);
  int nblocksx2=ceil(1.*n1/BLOCK_SIZE);
  int nblocksy2=ceil(1.*n2/BLOCK_SIZE);
  dim3 dimBlock(BLOCK_SIZE, BLOCK_SIZE);
  dim3 dimGrid(nblocksx,nblocksy);
  dim3 dimGrid2(nblocksx2,nblocksy2);

  for(int isou=(*tid); isou<sou_nx; isou+=(*nth))
  {
    sou_x=isou*sou_dx+sou_ox;

    cuda_call(cudaMemset(p0,   0,n1e*n2e*sizeof(float)));
    cuda_call(cudaMemset(p1,   0,n1e*n2e*sizeof(float)));
    cuda_call(cudaMemset(p2,   0,n1e*n2e*sizeof(float)));
    cuda_call(cudaMemset(p3,   0,n1e*n2e*sizeof(float)));
    cuda_call(cudaMemset(p4,   0,n1e*n2e*sizeof(float)));
    cuda_call(cudaMemset(p5,   0,n1e*n2e*sizeof(float)));
    cuda_call(cudaMemset(p6,   0,n1e*n2e*sizeof(float)));
    cuda_call(cudaMemset(p7,   0,n1e*n2e*sizeof(float)));
    cuda_call(cudaMemset(p8,   0,n1e*n2e*sizeof(float)));
    cuda_call(cudaMemset(p9,   0,n1e*n2e*sizeof(float)));
    cuda_call(cudaMemset(pinj0, 0,n1e*n2e*sizeof(float)));
    cuda_call(cudaMemset(pinj1, 0,n1e*n2e*sizeof(float)));
    cuda_call(cudaMemset(pinj2, 0,n1e*n2e*sizeof(float)));
    cuda_call(cudaMemset(pinj3, 0,n1e*n2e*sizeof(float)));
    cuda_call(cudaMemset(pinj4, 0,n1e*n2e*sizeof(float)));
    cuda_call(cudaMemset(data, 0,nt*rec_nx*sizeof(float)));

    its = 0;
    for(int it=0; it<n3; it++)
    {
      if(it%wfld_j==0)
      {
        kernel_exec(inject_source_kernel<<<1,1>>>(sourcefd, pinj0, sou_x, its));
        kernel_exec(escat_for_kernel<<<dimGrid2, dimBlock>>>(p1, refl, pinj1));
        kernel_exec(scat_kernel<<<dimGrid2, dimBlock>>>(p1, deltaB, pinj2));
        kernel_exec(scat_kernel<<<dimGrid2, dimBlock>>>(p3, deltaB, pinj3));
        kernel_exec(escat_for_kernel<<<dimGrid2, dimBlock>>>(p5, refl, pinj4));

        kernel_exec(wave_inj_kernel<<<dimGrid, dimBlock>>>(p0, p1, p0, pinj0));
        kernel_exec(wave_inj_kernel<<<dimGrid, dimBlock>>>(p2, p3, p2, pinj1));
        kernel_exec(wave_inj_kernel<<<dimGrid, dimBlock>>>(p4, p5, p4, pinj2));
        kernel_exec(wave_inj_kernel<<<dimGrid, dimBlock>>>(p6, p7, p6, pinj3));
        kernel_exec(wave_inj_kernel<<<dimGrid, dimBlock>>>(p8, p9, p8, pinj4));
      }
      else
      {
        kernel_exec(wave_kernel<<<dimGrid, dimBlock>>>(p0, p1, p0));
        kernel_exec(wave_kernel<<<dimGrid, dimBlock>>>(p2, p3, p2));
        kernel_exec(wave_kernel<<<dimGrid, dimBlock>>>(p4, p5, p4));
        kernel_exec(wave_kernel<<<dimGrid, dimBlock>>>(p6, p7, p6));
        kernel_exec(wave_kernel<<<dimGrid, dimBlock>>>(p8, p9, p8));
      }

      kernel_exec(fat_update1_kernel<<<n2e/80+1,80>>>(p1, p0));
      kernel_exec(fat_update2_kernel<<<n1e/80+1,80>>>(p1, p0));
      kernel_exec(fat_update1_kernel<<<n2e/80+1,80>>>(p3, p2));
      kernel_exec(fat_update2_kernel<<<n1e/80+1,80>>>(p3, p2));
      kernel_exec(fat_update1_kernel<<<n2e/80+1,80>>>(p5, p4));
      kernel_exec(fat_update2_kernel<<<n1e/80+1,80>>>(p5, p4));
      kernel_exec(fat_update1_kernel<<<n2e/80+1,80>>>(p7, p6));
      kernel_exec(fat_update2_kernel<<<n1e/80+1,80>>>(p7, p6));
      kernel_exec(fat_update1_kernel<<<n2e/80+1,80>>>(p9, p8));
      kernel_exec(fat_update2_kernel<<<n1e/80+1,80>>>(p9, p8));

      if(it%wfld_j==0)
      {
        kernel_exec(collect_data_kernel<<<rec_nx/80+1,80>>>(data, p7, its));
        kernel_exec(collect_data_kernel<<<rec_nx/80+1,80>>>(data, p9, its));
        its++;
      }

      ptemp=p0; p0=p1; p1=ptemp;
      ptemp=p2; p2=p3; p3=ptemp;
      ptemp=p4; p4=p5; p5=ptemp;
      ptemp=p6; p6=p7; p7=ptemp;
      ptemp=p8; p8=p9; p9=ptemp;
    }
    cuda_call(cudaMemcpy(&data_cpu[nt*rec_nx*isou], data, nt*rec_nx*sizeof(float), cudaMemcpyDeviceToHost));
  }
}

extern "C" void gpu_tomoadj(int *tid, int *nth, float *deltaB_cpu, float *data_cpu)
{

  int nt, n1, n2, n3;
  int h_nx, n1e, n2e, wfld_j;
  int sou_nx, sou_ox, sou_dx, rec_nx;
  float *p0, *p1, *p2, *p3;
  float *pinj0, *pinj1;
  float *sourcefd, *refl, *deltaB, *data, *wfld1, *wfld2;
  int sou_x, its;
  float *ptemp;

  cuda_call(cudaMemcpyFromSymbol(&nt, nt_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&n1, n1_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&n2, n2_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&n3, n3_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&h_nx, h_nx_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&n1e, n1e_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&n2e, n2e_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&wfld_j, wfld_j_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&sou_nx, sou_nx_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&sou_ox, sou_ox_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&sou_dx, sou_dx_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&rec_nx, rec_nx_gpu, sizeof(int)));
  cuda_call(cudaMemcpyFromSymbol(&p0, p0_gpu, sizeof(float*)));
  cuda_call(cudaMemcpyFromSymbol(&p1, p1_gpu, sizeof(float*)));
  cuda_call(cudaMemcpyFromSymbol(&p2, p2_gpu, sizeof(float*)));
  cuda_call(cudaMemcpyFromSymbol(&p3, p3_gpu, sizeof(float*)));
  cuda_call(cudaMemcpyFromSymbol(&pinj0, pinj0_gpu, sizeof(float*)));
  cuda_call(cudaMemcpyFromSymbol(&pinj1, pinj1_gpu, sizeof(float*)));
  cuda_call(cudaMemcpyFromSymbol(&sourcefd, sourcefd_gpu, sizeof(float*)));
  cuda_call(cudaMemcpyFromSymbol(&refl, refl_gpu, sizeof(float*)));
  cuda_call(cudaMemcpyFromSymbol(&deltaB, deltaB_gpu, sizeof(float*)));
  cuda_call(cudaMemcpyFromSymbol(&data, data_gpu, sizeof(float*)));
  cuda_call(cudaMemcpyFromSymbol(&wfld1, wfld1_gpu, sizeof(float*)));
  cuda_call(cudaMemcpyFromSymbol(&wfld2, wfld2_gpu, sizeof(float*)));

  int nblocksx=ceil(1.*(n1e-2*FAT)/BLOCK_SIZE);
  int nblocksy=ceil(1.*(n2e-2*FAT)/BLOCK_SIZE);
  int nblocksx2=ceil(1.*n1/BLOCK_SIZE);
  int nblocksy2=ceil(1.*n2/BLOCK_SIZE);
  dim3 dimBlock(BLOCK_SIZE, BLOCK_SIZE);
  dim3 dimGrid(nblocksx,nblocksy);
  dim3 dimGrid2(nblocksx2,nblocksy2);

  cuda_call(cudaMemset(deltaB,   0,n1*n2*sizeof(float)));

  for(int isou=(*tid); isou<sou_nx; isou+=(*nth))
  {
    sou_x=isou*sou_dx+sou_ox;

    cuda_call(cudaMemset(p0,   0,n1e*n2e*sizeof(float)));
    cuda_call(cudaMemset(p1,   0,n1e*n2e*sizeof(float)));
    cuda_call(cudaMemset(p2,   0,n1e*n2e*sizeof(float)));
    cuda_call(cudaMemset(p3,   0,n1e*n2e*sizeof(float)));
    cuda_call(cudaMemset(pinj0, 0,n1e*n2e*sizeof(float)));
    cuda_call(cudaMemset(pinj1, 0,n1e*n2e*sizeof(float)));
    cuda_call(cudaMemset(wfld1, 0,n1*n2*nt*sizeof(float)));
    cuda_call(cudaMemset(wfld2, 0,n1*n2*nt*sizeof(float)));
    cuda_call(cudaMemset(data, 0,nt*rec_nx*sizeof(float)));

    cuda_call(cudaMemcpy(data, &data_cpu[nt*rec_nx*isou], nt*rec_nx*sizeof(float), cudaMemcpyHostToDevice));

    its = 0;
    for(int it=0; it<n3; it++)
    {
      if(it%wfld_j==0)
      {
        kernel_exec(inject_source_kernel<<<1,1>>>(sourcefd, pinj0, sou_x, its));
        kernel_exec(escat_for_kernel<<<dimGrid2, dimBlock>>>(p1, refl, pinj1));
        kernel_exec(wave_inj_kernel<<<dimGrid, dimBlock>>>(p0, p1, p0, pinj0));
        kernel_exec(wave_inj_kernel<<<dimGrid, dimBlock>>>(p2, p3, p2, pinj1));
      }
      else
      {
        kernel_exec(wave_kernel<<<dimGrid, dimBlock>>>(p0, p1, p0));
        kernel_exec(wave_kernel<<<dimGrid, dimBlock>>>(p2, p3, p2));
      }
      kernel_exec(fat_update1_kernel<<<n2e/80+1,80>>>(p1, p0));
      kernel_exec(fat_update2_kernel<<<n1e/80+1,80>>>(p1, p0));
      kernel_exec(fat_update1_kernel<<<n2e/80+1,80>>>(p3, p2));
      kernel_exec(fat_update2_kernel<<<n1e/80+1,80>>>(p3, p2));
      if(it%wfld_j==0)
      {
        kernel_exec(collect_wfld_kernel<<<dimGrid2, dimBlock>>>(wfld1, p1, its));
        kernel_exec(collect_wfld_kernel<<<dimGrid2, dimBlock>>>(wfld2, p3, its));
        its++;
      }
      ptemp=p0; p0=p1; p1=ptemp;
      ptemp=p2; p2=p3; p3=ptemp;
    }

    cuda_call(cudaMemset(p0,   0,n1e*n2e*sizeof(float)));
    cuda_call(cudaMemset(p1,   0,n1e*n2e*sizeof(float)));
    cuda_call(cudaMemset(p2,   0,n1e*n2e*sizeof(float)));
    cuda_call(cudaMemset(p3,   0,n1e*n2e*sizeof(float)));
    cuda_call(cudaMemset(pinj0, 0,n1e*n2e*sizeof(float)));
    cuda_call(cudaMemset(pinj1, 0,n1e*n2e*sizeof(float)));

    its = nt-1;
    for(int it=n3-1; it>=0; it--)
    {
      if(it%wfld_j==0)
      {
        kernel_exec(inject_data_kernel<<<rec_nx/80+1,80>>>(data, pinj0, its));
        kernel_exec(escat_adj_kernel<<<dimGrid2, dimBlock>>>(p1, refl, pinj1));
        kernel_exec(wave_adj_inj_kernel<<<dimGrid, dimBlock>>>(p0, p1, p0, pinj0));
        kernel_exec(wave_adj_inj_kernel<<<dimGrid, dimBlock>>>(p2, p3, p2, pinj1));
      }
      else
      {
        kernel_exec(wave_adj_kernel<<<dimGrid, dimBlock>>>(p0, p1, p0));
        kernel_exec(wave_adj_kernel<<<dimGrid, dimBlock>>>(p2, p3, p2));
      }
      kernel_exec(fat_update1_kernel<<<n2e/80+1,80>>>(p1, p0));
      kernel_exec(fat_update2_kernel<<<n1e/80+1,80>>>(p1, p0));
      kernel_exec(fat_update1_kernel<<<n2e/80+1,80>>>(p3, p2));
      kernel_exec(fat_update2_kernel<<<n1e/80+1,80>>>(p3, p2));
      if(it%wfld_j==0)
      {
        kernel_exec(image_kernel<<<dimGrid2, dimBlock>>>(p3, deltaB, wfld1, its));
        kernel_exec(image_kernel<<<dimGrid2, dimBlock>>>(p1, deltaB, wfld2, its));
        its--;
      }
      ptemp=p0; p0=p1; p1=ptemp;
      ptemp=p2; p2=p3; p3=ptemp;
    }
  }
  cuda_call(cudaMemcpy(deltaB_cpu, deltaB, n1*n2*sizeof(float), cudaMemcpyDeviceToHost));
}

